# -*- coding: utf-8 -*-
# @Time    : 2023/4/2 15:17
# @Author  : luyi
import inspect
from typing import Tuple, Optional, Any
from math import ceil, floor
from .constants import Vtype

INFINITY = 1000000000000


def get_update_lub(is_cp, lb, ub: Any, vtype: Vtype):
    if is_cp and ub is None:
        ub = INFINITY
    elif is_cp and ub:
        ...
    elif ub is None:
        ub = INFINITY
    if is_cp:
        if lb:
            lb = ceil(lb)
        if ub:
            ub = floor(ub)
    if vtype == Vtype.BINARY:
        if ub > 0 or ub is None:
            ub = 1
        if lb > 0:
            lb = 1
        if lb < 0 or lb is None:
            lb = 0
    ub = max(ub, 1)
    if lb > ub:
        raise Exception("0-1变量lb>=ub")
    return lb, ub


def get_length(arg) -> int:
    if is_list_or_tuple(arg):
        return len(arg)
    return 1


def is_list_or_tuple(items) -> bool:
    """
    是否是list或tupe类型
    """
    return isinstance(items, (list, tuple))


def is_real(expr):
    return isinstance(expr, (int, float)) and not isinstance(expr, (complex,))


def is_generator(obj):
    return inspect.isgeneratorfunction(obj) or inspect.isgenerator(obj)


def get_combinations(lst):
    """
    对于一个由若干个子列表组成的列表，返回其中所有元素的组合方式
    例如，对于输入 [[1,2],[3,4]] ，调用 get_combinations() 函数会返回 [[1, 3], [1, 4], [2, 3], [2, 4]]，
    对于更复杂的列表，例如 [[1, 2], [3], [4, 5, 6]] ，调用 get_combinations() 函数会返回
    [[1, 3, 4], [1, 3, 5], [1, 3, 6], [2, 3, 4], [2, 3, 5], [2, 3, 6]]，
    也就是每个子列表中的元素都和其它子列表中的所有元素进行了组合。
    :param lst: 由若干个子列表组成的列表
    :return: 所有元素的组合方式，以列表形式返回
    """
    if not isinstance(lst, list):
        lst = [lst]
    if len(lst) == 1:
        return lst[0]
    else:
        combinations = []
        sub_combinations = get_combinations(lst[1:])
        if not isinstance(sub_combinations, list):
            sub_combinations = [sub_combinations]
        elems = lst[0]
        if not isinstance(elems, list):
            elems = [elems]
        for elem in elems:
            for sub_elem in sub_combinations:
                com = ()
                if is_list_or_tuple(elem):
                    elem = tuple(elem)
                else:
                    elem = tuple([elem])
                if is_list_or_tuple(sub_elem):
                    sub_elem = tuple(sub_elem)
                else:
                    sub_elem = tuple([sub_elem])
                com += elem + sub_elem
                combinations.append(com)
        return combinations


def tuple_fuzzy_match(tuple_value1, tuple_value2) -> bool:
    if not is_list_or_tuple(tuple_value1) and not is_list_or_tuple(tuple_value2):
        tuple_value1 = [tuple_value1]
        tuple_value2 = [tuple_value2]
    if len(tuple_value1) != len(tuple_value2):
        raise ValueError("元组大小不匹配")
    for i, v in enumerate(tuple_value1):
        if v == "*" or tuple_value2[i] == "*":
            continue
        if v != tuple_value2[i]:
            return False
    return True


def tuple_fuzz_match_list(
    tuple_value1, tuple_value_list
) -> Tuple[bool, Optional[tuple]]:
    """
    模糊匹配
    :param tuple_value1:
    :param tuple_value_list:
    :return:
    """
    for v in tuple_value_list:
        if tuple_fuzzy_match(tuple_value1, v):
            return True, v
    return False, None


def check_bool_var(*args):
    """检测若不是ool变量，则抛出异常"""
    for arg in args:
        if is_list_or_tuple(arg) or is_generator(arg):
            for ar in arg:
                if not is_bool_var(ar):
                    raise RuntimeError(f"{ar.VarName} is not a binary variable.")
        else:
            if not is_bool_var(arg):
                raise RuntimeError(f"{arg.VarName} is not a binary variable.")


def is_bool_var(var) -> bool:
    """是否是bool变量"""
    try:
        if not (var.UB == 1 and var.LB == 0):
            return False
        return True
    except Exception as _:
        return True


def pre_condition(condition: bool, msg: str):
    if not condition:
        raise Exception(msg)


if __name__ == "__main__":
    print(get_combinations([[(2, 8), (2, 3)], [(1, 2), (2, 3)]]))



# -*- coding: utf-8 -*-
# @Time    : 2023/4/7 8:07
# @Author  : luyi
"""
变量
"""
from ortools.linear_solver.pywraplp import Variable
from ortools.sat.python.cp_model import IntVar, CpSolver
from ortools.util.python.sorted_interval_list import Domain
from .interface_ import *


class IntervalVar:
    def __init__(self, start, size, end) -> None:
        self.start = start
        self.size = size
        self.end = end


class Var(Variable, IVar):

    # def Not(self) -> "IVar":
    #     if self.v_type == Vtype.BINARY or self.v_type == Vtype.INTEGER:
    #         z = self._solver.IntVar(  # type: ignore
    #             lb=0, ub=1, name=f"{self.VarName}.Not"
    #         )  # type: ignore
    #         self._solver.Add(z + self == 1)  # type: ignore
    #         return z
    #     else:
    #         raise RuntimeError("Only for binary variable .")  # type: ignore

    # def setUb(self, ub):
    #     self.SetUb(ub)

    # def setLb(self, lb):
    #     self.SetLb(lb)

    # def setBounds(self, lb, ub):
    #     self.SetBounds(lb, ub)

    # def setValue(self, value):
    #     self.SetBounds(value, value)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.v_type = None

    @property
    def X(self):
        return self.solution_value()

    # @property
    # def VarIndex(self):
    #     return self.index()

    @property
    def VarName(self) -> str:
        return self.name()

    @property
    def LB(self):
        return self.lb()

    @property
    def UB(self):
        return self.ub()


Variable.X = Var.X  # type: ignore
Variable.VarName = Var.VarName  # type: ignore
# Variable.VarIndex = Var.VarIndex  # type: ignore
Variable.LB = Var.LB  # type: ignore
Variable.UB = Var.UB  # type: ignore
# Variable.setValue = Var.setValue  # type: ignore
# Variable.setLb = Var.setLb  # type: ignore
# Variable.setUb = Var.setUb  # type: ignore
# Variable.setBounds = Var.setBounds  # type: ignore
# Variable.Not = Var.Not  # type: ignore


class CpVar(IntVar, IVar):
    def __init__(self, model, domain, is_bool: bool, name, solver):
        self._solver: CpSolver = solver
        self.__domain: Domain = domain
        try:
            super().__init__(model, domain, is_bool, name)  # type: ignore
        except:
            super().__init__(model, domain, name)  # type: ignore

    # def Not(self):  # type: ignore
    #     y = super().Not()
    #     y.VarName = f"{self.VarName}.not"  # type: ignore
    #     y.X = lambda: self._solver.valueExpression(y)  # type: ignore
    #     return y

    # def setValue(self, value):
    #     raise Exception("not impl")

    # def setUb(self, ub):
    #     raise Exception("not impl")

    # def setLb(self, lb):
    #     raise Exception("not impl")

    # def setBounds(self, lb, ub):
    #     raise Exception("not impl")

    @property
    def X(self):
        return self._solver.valueExpression(self)  # type: ignore

    # @property
    # def VarIndex(self):
    #     return self.Index()

    @property
    def VarName(self):
        return self.Name()

    @property
    def LB(self):
        try:
            return self.__domain.Min()  # type: ignore
        except:
            return self.__domain.min()

    @property
    def UB(self):
        try:
            return self.__domain.Max()  # type: ignore
        except:
            return self.__domain.max()



import pytest
import mipx

SOLVERS = ["CBC", "SCIP", "GUROBI", "CPLEX", "SAT", "CP", "CPLEX_CP"]
CP_SOLVERS = ["CP", "CPLEX_CP"]


@pytest.fixture(params=SOLVERS)
def model(request):
    if request.param in CP_SOLVERS:
        return mipx.CpModel(solver_id=request.param)
    return mipx.Model(solver_id=request.param)


@pytest.fixture(params=CP_SOLVERS)
def cpmodel(request):
    return mipx.CpModel(solver_id=request.param)


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
@pytest.mark.usefixtures("model")
class TestModel:
    def test_add_constr(self, model: mipx.Model):
        x = model.addVar(ub=1)
        y = model.addVars(10)
        model.addConstr(x == 1)
        model.addConstr(y.quicksum() == 10)
        y.quicksum([1, 2])
        status = model.optimize()
        assert mipx.success(status)
        assert x.X == 1
        # assert model.valueExpression(y.quicksum()) == 10

    def test_gen_constr_and(self, model: mipx.Model):
        x = model.addVars(6, vtype=mipx.BINARY)
        y = model.addVar(vtype=mipx.BINARY)
        model.addConstr(x[0] == 0)
        model.addConstr(x.quicksum() == 5)
        model.addGenConstrAnd(y, x.quickselect())
        m = model.addVar(vtype=mipx.BINARY, ub=1)
        z = model.addVars(3, vtype=mipx.BINARY, ub=1)
        model.addConstr(z.quicksum() == 3)
        model.addGenConstrAnd(m, z.quickselect())
        status = model.optimize()
        assert mipx.success(status)
        assert y.X == 0
        assert m.X == 1

    def test_gen_constr_or(self, model: mipx.Model):
        x = model.addVars(6, vtype=mipx.BINARY)
        y = model.addVar(vtype=mipx.BINARY)
        model.addConstr(x[0] == 0)
        model.addConstr(x.quicksum() == 5)
        model.addGenConstrOr(y, x.quickselect())
        m = model.addVar(vtype=mipx.BINARY, ub=1)
        z = model.addVars(3, vtype=mipx.BINARY, ub=1)
        model.addConstr(z.quicksum() == 0)
        model.addGenConstrOr(m, z.quickselect())
        status = model.optimize()
        assert mipx.success(status)
        assert y.X == 1
        assert m.X == 0

    def test_gen_constr_xor(self, model: mipx.Model):
        x = model.addVars(2, vtype=mipx.BINARY)
        y = model.addVar(vtype=mipx.BINARY)
        # 不同时
        model.addConstrs([x[0] == 0, x[1] == 1])
        model.addGenConstrXOr(y, x.quickselect())
        a = model.addVars(2, vtype=mipx.BINARY)
        b = model.addVar(vtype=mipx.BINARY)
        # 同时
        model.addConstrs([a[0] == 1, a[1] == 1])
        model.addGenConstrXOr(b, a.quickselect())
        c = model.addVars(2, vtype=mipx.BINARY)
        d = model.addVar(vtype=mipx.BINARY)
        # 同时
        model.addConstrs([c[0] == 0, c[1] == 0])
        model.addGenConstrXOr(d, c.quickselect())
        e = model.addVars(2, vtype=mipx.BINARY)
        f = model.addVar(vtype=mipx.BINARY)
        # 同时
        model.addConstrs([e[0] == 1, e[1] == 0])
        model.addGenConstrXOr(f, e.quickselect())
        status = model.optimize()
        assert mipx.success(status)
        assert mipx.succ(status)
        assert y.X == 1
        assert b.X == 0
        assert d.X == 0
        assert f.X == 1

    def test_gen_constr_pwl(self, model: mipx.Model):
        x = model.addVar(ub=10, vtype=mipx.INTEGER)
        y = model.addVar(ub=100, vtype=mipx.INTEGER)
        z = model.addVar(lb=1, ub=10, vtype=mipx.INTEGER)
        model.addGenConstrPWL(x, y, [4, 10], [1, 5 + z, 10], M=1000)
        model.setObjective(y)
        status = model.optimize(mipx.MAXIMIZE)
        assert mipx.success(status)
        assert y.X == 15
        assert z.X == 10

    # def test_gen_constr_indicator(self, model: mipx.Model):
    #     x = model.addVar(ub=10, vtype=mipx.INTEGER)
    #     y = model.addVar(ub=100, vtype=mipx.INTEGER)
    #     z = model.addVar(lb=1, ub=10, vtype=mipx.INTEGER)
    #     model.addGenConstrPWL(x, y, [4, 10], [1, 5 + z, 10], M=1000)
    #     model.setObjective(y)
    #     status = model.optimize(mipx.MAXIMIZE)
    #     assert mipx.success(status)
    #     assert y.X == 15
    #     assert z.X == 10

    def test_gen_constr_abs(self, model: mipx.Model):
        x = model.addVar(lb=-10, ub=10, vtype=mipx.INTEGER)
        y = model.addVar(ub=10, vtype=mipx.INTEGER)
        model.addGenConstrAbs(y, x, M=1000)
        model.addConstr(x == -5)
        status = model.optimize(mipx.MAXIMIZE)
        assert mipx.success(status)
        assert y.X == 5



# -*- coding: utf-8 -*-
# @Time    : 2023/4/28 23:54
# @Author  : luyi

from collections import UserDict
from .constants import OptimizationStatus
from .utilx import is_list_or_tuple


def debugVar(vars, no_zero=False):
    """
    调试var的方法

    :param _type_ vars: 可以是变量的集合或者单个变量
    :param bool no_zero: 不输出<=0的值, defaults to True
    """
    is_list = False
    if isinstance(vars, UserDict) or isinstance(vars, dict):
        vars = vars.values()
        is_list = True
    if is_list_or_tuple(vars):
        is_list = True
    if is_list:
        for var in vars:
            if no_zero:
                if var.X <= 0.01:
                    continue
            print(f"{var.VarName}:{var.X}")
    else:
        print(f"{vars.VarName}:{vars.X}")  # type:ignore


def success(statue: OptimizationStatus):
    """
    判断是否求解成功

    :param statue: 求解状态
    :return: 是否成功
    """
    return statue in [OptimizationStatus.OPTIMAL, OptimizationStatus.FEASIBLE]


def name_str(base_name, *key) -> str:
    """
    名称组合
    """
    if base_name == "":
        return ""
    re_name = key
    if is_list_or_tuple(key):
        re_name = ",".join(str(x) for x in key)
    return f"{base_name}[{re_name}]"



# -*- coding: utf-8 -*-
# @Time    : 2023/3/31 22:04
# @Author  : luyi
from typing import Any, List, Optional, Union, Tuple, Dict
from typing_extensions import override
from ortools.linear_solver import pywraplp

try:
    from ortools.linear_solver.linear_solver_natural_api import OFFSET_KEY  # type: ignore
except Exception as e:
    from ortools.linear_solver.python.linear_solver_natural_api import OFFSET_KEY
from .constants import Vtype, ObjType, OptimizationStatus, CmpType
from .interface_ import IModel, IVar, ITupledict
from .tupledict import tupledict
from .utilx import (
    is_list_or_tuple,
    get_combinations,
    is_generator,
    is_bool_var,
    INFINITY,
)
from .func import name_str
from .variable import IntervalVar


Real = Union[float, int]


class LinearConstraint(pywraplp.LinearConstraint):
    """
    线性约束

    Args:
        pywraplp (_type_): 线性
    """

    def __init__(self, expr, lb, ub):
        self.__expr = expr
        self.__lb = lb
        self.__ub = ub
        super().__init__(expr, lb, ub)

    def lbAndUb(self) -> Tuple[float, float]:
        coeffs = self.__expr.GetCoeffs()
        constant = coeffs.pop(OFFSET_KEY, 0.0)
        lb = -INFINITY
        ub = INFINITY
        if self.__lb > -INFINITY:
            lb = self.__lb - constant
        if self.__ub < INFINITY:
            ub = self.__ub - constant
        return lb, ub

    def linear_expr(self) -> pywraplp.LinearExpr:
        return self.__expr


class CommonModelSolver(IModel):
    """
    模型
    """

    def __init__(self, solver_id="SCIP", name=""):
        super().__init__(solver_id, name)
        self._name = name
        self.__solver: pywraplp.Solver = pywraplp.Solver.CreateSolver(solver_id)
        self._objective_n_list = []
        self._flag_objective = False
        self._num_inter_vars = 0
        self._num_binary_vars = 0
        self._num_coutinuous_vars = 0

    @override
    def Sum(self, expr_array) -> float:
        result = pywraplp.SumArray(expr_array)
        return result  # type: ignore

    @override
    def setHint(self, start: Dict[IVar, Real]):
        _vars = []
        _values = []
        for var, value in start.items():
            _vars.append(var)
            _values.append(value)
        self.__solver.SetHint(_vars, _values)

    @override
    def setTimeLimit(self, time_limit_seconds):
        self.__solver.set_time_limit(time_limit_seconds * 1000)

    @override
    def wall_time(self) -> int:
        """
        求解所花费的时间 milliseconds
        :return:
        """
        return self.__solver.wall_time()

    @override
    def iterations(self) -> int:
        """迭代次数"""
        return self.__solver.iterations()

    @override
    def nodes(self) -> int:
        return self.__solver.nodes()

    @override
    def addVar(
        self,
        lb: Real = 0.0,
        ub: Real = INFINITY,
        vtype: Vtype = Vtype.CONTINUOUS,
        name: str = "",
    ) -> IVar:
        """
        Add a decision variable to a model.
        :param lb: Lower bound for new variable.
        :param ub: Upper bound for new variable.
        :param vtype: variable type for new variable(Vtype.CONTINUOUS, Vtype.BINARY, Vtype.INTEGER).
        :param name: Name for new variable.write
        :return: variable.
        """
        var: Optional[IVar] = None
        if vtype == Vtype.CONTINUOUS:
            self._num_coutinuous_vars += 1
            var = self.__solver.NumVar(lb, ub, name)
        elif vtype == Vtype.INTEGER:
            self._num_inter_vars += 1
            var = self.__solver.IntVar(lb, ub, name)
        elif vtype == Vtype.BINARY:
            self._num_binary_vars += 1
            var = self.__solver.IntVar(lb, ub, name)
        var.v_type = vtype  # type: ignore
        var._solver = self  # type: ignore
        return var  # type: ignore

    @override
    def addVars(
        self,
        *indices,
        lb: Real = 0.0,
        ub: Real = INFINITY,
        vtype: Vtype = Vtype.CONTINUOUS,
        name: str = "",
    ) -> ITupledict:
        li = []
        for ind in indices:
            if isinstance(ind, int):
                ind = [i for i in range(ind)]
            elif is_list_or_tuple(ind):
                pass
            elif isinstance(ind, str):
                ind = [ind]
            else:
                raise ValueError("error input")
            li.append(ind)
        all_keys_tuple = get_combinations(li)
        tu_dict = tupledict(
            [
                [key, self.addVar(lb, ub, vtype, name_str(name, key))]
                for key in all_keys_tuple
            ]
        )
        return tu_dict

    @override
    def getVars(self) -> List[IVar]:
        """
        Retrieve a list of all variables in the model.
        :return: All variables in the model.
        """
        return self.__solver.variables()

    @override
    def getVar(self, i) -> IVar:
        """
        获得第i个变量
        :param i:
        :return:
        """
        return self.__solver.variable(i)

    @override
    def addConstr(self, lin_expr):
        self.__solver.Add(lin_expr)

    @override
    def addConstrs(self, lin_exprs, name=""):
        if not is_list_or_tuple(lin_exprs) and not is_generator(lin_exprs):
            raise RuntimeError("constraint conditions are not a set or list")
        for i, lin_expr in enumerate(lin_exprs):
            self.addConstr(lin_expr)

    @override
    def setObjective(self, expr):
        """
        Set the model objective equal to a linear expression
        :param expr:New objective expression
        :param obj_type:Optimization sense (Sense.MINIMIZE for minimization, Sense.MAXIMIZE for maximization)
        """
        self._objective_n_list.append(expr)

    def _set_objective_sense(self, obj_type: ObjType):
        if obj_type == ObjType.MINIMIZE:
            self.__solver.Objective().SetMinimization()
        else:
            self.__solver.Objective().SetMaximization()

    @override
    def setObjectiveN(
        self, expr, index: int, priority: int = 0, weight: float = 1, name: str = ""
    ):
        """
        多目标优化，优化最小值
        :param name:
        :param expr: 表达式
        :param index: 目标函数对应的序号 (默认 0，1，2，…), 以 index=0 作为目标函数的值, 其余值需要另外设置参数
        :param priority:分层序列法多目标决策优先级(整数值), 值越大优先级越高,# 未实现。
        :param weight: 线性加权多目标决策权重(在优先级相同时发挥作用)
        """
        self._objective_n_list.append(expr * weight)

    @override
    def addGenConstrAnd(self, resvar, varList: List[IVar]):
        """
        和 addGenConstrAnd(y, [x1,x2])
        :param resvar:
        :param varList:
        :return:
        """
        super().addGenConstrAnd(resvar, varList)

    @override
    def addGenConstrOr(self, resvar: IVar, varList: List[IVar]):
        """
        或
        :param resvar:
        :param varList:
        :return:
        """
        super().addGenConstrOr(resvar, varList)

    @override
    def addGenConstrXOr(self, resvar: IVar, varList: List[IVar]):
        """
        异或
        :param resvar:
        :param varList:
        :return:
        """
        super().addGenConstrXOr(resvar, varList)

    @override
    def addGenConstrPWL(self, var_x, var_y, x_range, y_range, M=None):
        """
        设置分段约束
        :param var_y: f(x)
        :param var_x:指定变量的目标函数是分段线性
        :param x:  定义分段线性变量的点的范围边界(非减序列)
        :param y:定义分段线性变量的范围所对应的目标函数的值
        :return:
        """
        super().addGenConstrPWL(var_x, var_y, x_range, y_range, M)

    @override
    def addGenConstrIndicator(
        self,
        binvar: IVar,
        binval: bool,
        lhs: IVar,
        sense: CmpType,
        rhs: float,
        M: float,
    ):
        """
        若 binvar 为binval ,则 lhs 与 rhs 之间有sense 的关系

        :param binvar: 0-1变量
        :param binval: bool 常量
        :param lhs:  左侧变量
        :param sense: 等号，大于等于，小于等于
        :param rhs: 右侧常量
        :param M: 大M
        :return:
        """
        super().addGenConstrIndicator(binvar, binval, lhs, sense, rhs, M)

    @override
    def addIndicator(self, binvar: IVar, binval: bool, constr):
        raise RuntimeError("not implemented")

    @override
    def addGenConstrAbs(self, resvar, var_abs: IVar, M=None):  # type: ignore
        x = self.addVars(2, name="")
        _M = 10000000
        if M is not None:
            _M = M
        self.addConstr(var_abs == x[0] - x[1])
        self.addConstr(resvar == x[0] + x[1])
        y = self.addVar(vtype=Vtype.BINARY)
        # x[0] * x[1]==0 的约束
        self.addConstr(x[0] <= y * _M)
        self.addConstr(x[1] <= (1 - y) * _M)

    @override
    def addGenConstrMultiply(self, z: IVar, l: Tuple[IVar, IVar]):
        """x * y = z"""
        super().addGenConstrMultiply(z, l)
        x = l[0]
        y = l[1]
        if not is_bool_var(x) and not is_bool_var(y):
            raise RuntimeError("At least one binary variable is required.")
        if is_bool_var(y):
            x, y = y, x
        M = y.UB
        if self.solver_id in ["SAT", "CBC"]:
            M = INFINITY if M == INFINITY else M
        self.addConstr(z <= y)
        self.addConstr(z <= x * M)
        self.addConstr(z >= y + (x - 1) * M)

    @override
    def addRange(
        self, expr, min_value: Union[float, int], max_value: Union[float, int]
    ):
        """
        添加范围约束
        :param
        expr:
        :param
        min_value:
        :param
        max_value:
        :param
        :return:
        """
        super().addRange(expr, min_value, max_value)

    @override
    def calc_lbub_and_expr(self, constr) -> Tuple[Real, Real, Any]:
        tempConstr = _change_to_linear_constraint(constr)
        lb, ub = tempConstr.lbAndUb()
        expr = tempConstr.linear_expr()
        return lb, ub, expr

    @override
    def addConstrOr(
        self,
        constrs: List,
        ok_num: int = 1,
        cmp_type: CmpType = CmpType.EQUAL,
        M=None,
    ):
        """
            约束的满足情况

        :param constr: 所有的约束
        :param ok_num: 需要满足的个数，具体则根据cmpType
        :param cmpType: CmpType.LESS_EQUAL CmpType.EQUAL,CmpType.GREATER_EQUAL
        :return:
        """
        constr_num = len(constrs)
        x = []
        for i in range(constr_num):
            x.append(self.addVar(vtype=Vtype.BINARY))
            tempM = 0
            if M is not None:
                tempM = M
            lb, ub, expr = self.calc_lbub_and_expr(constrs[i])
            if lb > -INFINITY:  # 若大于
                if M is None:
                    tempM = (abs(lb) + 1) * 10
                self.addConstr(expr + tempM * (1 - x[i]) >= lb)
            if ub < INFINITY:
                if M is None:
                    tempM = (abs(ub) + 1) * 10
                self.addConstr(expr - tempM * (1 - x[i]) <= ub)
        if cmp_type == CmpType.EQUAL:
            self.addConstr(self.Sum(x) == ok_num)
        elif cmp_type == CmpType.GREATER_EQUAL:
            self.addConstr(self.Sum(x) >= ok_num)
        elif cmp_type == CmpType.LESS_EQUAL:
            self.addConstr(self.Sum(x) <= ok_num)  # type: ignore
        else:
            raise Exception("error value of cmpType")

    @override
    def numVars(self, vtype: Optional[Vtype] = None) -> int:
        """
        当前的变量的个数
        :return:
        """
        if vtype is Vtype.CONTINUOUS:
            return self._num_coutinuous_vars
        elif vtype is Vtype.INTEGER:
            return self._num_inter_vars
        elif vtype is Vtype.BINARY:
            return self._num_binary_vars
        else:
            return self.__solver.NumVariables()

    @override
    def numConstraints(self) -> int:
        """
        当前的约束的个数
        :return:
        """
        return self.__solver.NumConstraints()

    @override
    def write(self, filename: str, obfuscated=False):
        """
        写入到文件
        :param filename:文件名，支持后缀 .lp .mps .proto
        :param obfuscated:
        :return:
        """
        filename = filename.lower()
        content = ""
        if filename.endswith(".lp"):
            content = self.__solver.ExportModelAsLpFormat(False)
        elif filename.endswith(".mps"):
            content = self.__solver.ExportModelAsMpsFormat(True, obfuscated)
        elif filename.endswith(".proto"):
            raise TypeError(".proto 导出异常，待修复")

        with open(filename, "w", encoding="utf-8") as f:
            f.write(content)

    @override
    def read(self, path: str):
        raise RuntimeError("not implemented")

    @property
    @override
    def ObjVal(self):
        return self.__solver.Objective().Value()

    def _set_params(self):
        if self._params.TimeLimit:
            self.setTimeLimit(self._params.TimeLimit)
            # 是否允许输出运算信息，包括gap等
        if self._params.EnableOutput:
            self.__solver.EnableOutput()

    @override
    def optimize(self, obj_type: ObjType = ObjType.MINIMIZE) -> OptimizationStatus:
        self._set_params()
        # GAP 设置。
        params = pywraplp.MPSolverParameters()
        # 配置参数
        if self._params.MIPGap:
            params.SetDoubleParam(params.RELATIVE_MIP_GAP, self._params.MIPGap)
        if self._params.Precision:
            params.SetDoubleParam(params.PRIMAL_TOLERANCE, self._params.Precision)
        if obj_type == ObjType.MINIMIZE:
            self.__solver.Minimize(self.Sum(self._objective_n_list))
        else:
            self.__solver.Maximize(self.Sum(self._objective_n_list))
        status = self.__solver.Solve(params)

        if status == pywraplp.Solver.OPTIMAL:
            result = OptimizationStatus.OPTIMAL
        elif status == pywraplp.Solver.INFEASIBLE:
            result = OptimizationStatus.INFEASIBLE
        elif status == pywraplp.Solver.UNBOUNDED:
            result = OptimizationStatus.UNBOUNDED
        elif status == pywraplp.Solver.FEASIBLE:
            result = OptimizationStatus.FEASIBLE
        elif status == pywraplp.Solver.NOT_SOLVED:
            result = OptimizationStatus.NO_SOLUTION_FOUND
        else:
            result = OptimizationStatus.ERROR
        return result

    @override
    def clear(self):
        self.__solver.Clear()

    @override
    def close(self):
        self.clear()
        self.__solver = None  # type: ignore

    @override
    def valueExpression(self, expression):
        return expression.solution_value()
        # value = 0
        # coeffs = expression.GetCoeffs()
        # for key, coeff in coeffs.items():
        #     if key is OFFSET_KEY:
        #         value += coeffs.get(key)
        #     else:
        #         var: IVar = key
        #         x = var.X
        #         value += coeff * x
        # return value

    @override
    def newIntervalVar(self, start, size, end) -> IntervalVar:
        self.addConstr(start + size == end)
        return IntervalVar(start, size, end)

    @override
    def addNoOverlap(self, interval_vars: List[IntervalVar], M: int):
        m = INFINITY if M is None else M
        for i, var_i in enumerate(interval_vars):
            for j, var_j in enumerate(interval_vars):
                if i == j:
                    continue
                t = self.addVar(vtype=Vtype.BINARY)
                self.addConstr(var_i.start >= var_j.end - (1 - t) * m)
                self.addConstr(var_j.start >= var_i.end - t * m)

    @override
    def setNumThreads(self, num_theads: int):
        self.__solver.SetNumThreads(num_theads)


def _change_to_linear_constraint(
    constraint: pywraplp.LinearConstraint,
) -> LinearConstraint:
    expr = constraint._LinearConstraint__expr  # type: ignore
    ub = constraint._LinearConstraint__ub  # type: ignore
    lb = constraint._LinearConstraint__lb  # type: ignore
    return LinearConstraint(expr, lb, ub)



# -*- coding: utf-8 -*-
# @Time    : 2023/4/28 15:33
# @Author  : luyi
# 定义接口。统一类型。
from abc import ABCMeta, abstractmethod
from collections import UserDict, UserList
from typing import (
    Any,
    List,
    Literal,
    Optional,
    Union,
    Dict,
    Generator,
    Tuple,
)
from typing_extensions import TypeVarTuple, Generic, Unpack
from mipx.utilx import INFINITY, check_bool_var
from .constants import Vtype, ObjType, CmpType, OptimizationStatus, Params

Real = Union[float, int]


class IVar:
    def __init__(self):
        # 变量类型
        self.v_type = None
        self._solver = None

    @property
    def X(self) -> Real: ...

    # @property
    # def VarIndex(self) -> Real: ...

    @property
    def VarName(self) -> str: ...

    @property
    def LB(self) -> Real: ...

    @property
    def UB(self) -> Real: ...

    # def setValue(self, value: Real): ...

    # def setUb(self, ub: Real): ...

    # def setLb(self, lb: Real): ...

    # def Not(self) -> "IVar":
    #     """
    #     返回一个相反的变量，仅用于0,1变量。
    #     :return:
    #     """
    #     ...

    # def setBounds(self, lb, ub): ...

    def __add__(self, expr) -> Any: ...

    def __radd__(self, cst) -> Any: ...

    def __sub__(self, expr) -> Any: ...

    def __rsub__(self, cst) -> Any: ...

    def __mul__(self, other) -> Any: ...

    def __rmul__(self, other) -> Any: ...

    def __ge__(self, other) -> Any: ...

    def __le__(self, other) -> Any: ...


K = TypeVarTuple("K")


class ITupledict(UserDict, Generic[Unpack[K]]):
    """
    Custom Gurobi class: tupledict is a subclass of dict where
      the keys are a tuplelist.
    """

    @abstractmethod
    def __setitem__(self, key: tuple[Unpack[K]], item) -> None:
        super().__setitem__(key, item)

    @abstractmethod
    def prod(self, coeff: Dict, *pattern): ...

    @abstractmethod
    def select(self, *pattern) -> List: ...

    @abstractmethod
    def sum(self, *pattern) -> IVar: ...

    @abstractmethod
    def quicksum(self, *pattern) -> IVar: ...

    @abstractmethod
    def quickselect(self, *pattern) -> List[Any]: ...

    @abstractmethod
    def quickprod(self, coeff: Dict, *pattern) -> IVar: ...


class ITuplelist(UserList):
    @abstractmethod
    def select(self, *args) -> List: ...


class IModel(metaclass=ABCMeta):
    def __init__(
        self,
        solver_id: Literal[
            "SCIP",
            "CBC",
            "CP",
            "SAT",
            "CLP",
            "CPLEX",
            "CPLEX_CP",
            "GLOP",
            "BOP",
            "GLPK",
            "GUROBI",
        ],
        name,
    ):
        self._params: Params = Params()
        self.solver_id = solver_id
        self._name = name
        self._allkpis = {}
        self._multi_object_expr = []
        self._flag_objective_n = None

    @property
    def Params(self) -> Params:
        return self._params

    @property
    def INFINITY(self) -> Real:
        return INFINITY

    @property
    def name(self) -> str:
        return self._name

    @property
    def WallTime(self) -> float:
        return self.wall_time()

    def addKpi(self, kpi_arg, name: str):
        if name in self._allkpis:
            raise Exception(f"KPI名称 {name} 重复")
        self._allkpis[name] = kpi_arg

    def reportKpis(self):
        """
        输出所有KPI
        :return:
        """
        print("===========all kpis=============")
        for name, kpi_arg in self._allkpis.items():
            value = self.valueExpression(kpi_arg)
            print(f"{name} = {value}")
        print("================================")

    def kpiByName(self, name: str):
        """根据名称获取KPI"""
        kpi = self._allkpis.get(name)
        if kpi is None:
            raise Exception(f"KPI名称 {name} 不存在")
        return kpi

    def kpiValueByName(self, name: str):
        """根据名称获取KPI的值"""
        kpi = self.kpiByName(name)
        return self.valueExpression(kpi)

    def removeKpi(self, name: str):
        """
        移除KPI
        :param name:
        :return:
        """
        if name not in self._allkpis:
            raise Exception(f"KPI名称 {name} 不存在")
        del self._allkpis[name]

    def clearKpis(self):
        """
        清除所有KPI
        :return:
        """
        self._allkpis.clear()

    @property
    def numOfKpis(self):
        return len(self._allkpis)

    def statistics(self):
        """
        统计模型信息
        :return:
        """
        print("statistics：\n========")
        print(f"numConstraints：{self.numConstraints()}")
        print(f"numVars：{self.numVars()}")
        print(
            f"numVars of int：{self.numVars(Vtype.INTEGER)+self.numVars(Vtype.BINARY)}"
        )
        print("========")

    def reportMultiObjValue(self):
        """
        展示多目标优化的目标值
        """
        print("========MultiObjValue===========")
        for expr, weight, name in self._multi_object_expr:
            value = self.valueExpression(expr)
            print(f"{name} : {round(value*weight, 2)}")
        print("================================")

    # -----------------------抽象方法-------------------------------
    @abstractmethod
    def calc_lbub_and_expr(self, constr) -> Tuple[Real, Real, Any]:
        """
        将线性约束分解为变量下界，变量上界，表达式。

        :param constr: 线性约束
        """
        ...

    @abstractmethod
    def Sum(self, expr_array) -> Any: ...

    def sum(self, expr_array):
        return self.Sum(expr_array)

    @abstractmethod
    def setHint(self, start: Dict[IVar, Real]): ...

    @abstractmethod
    def setTimeLimit(self, time_limit_seconds: int):
        """
        设置程序最大运行时间
        :param time_limit_seconds: 秒
        """
        ...

    @abstractmethod
    def wall_time(self) -> Real:
        """
        求解所花费的时间
        :return: 求解所花费的时间，单位毫秒
        """
        ...

    @abstractmethod
    def iterations(self) -> Real:
        """

        :return: 算法迭代的次数
        """
        ...

    @abstractmethod
    def nodes(self) -> Real:
        """

        :return: 节点数
        """
        ...

    @abstractmethod
    def addVar(
        self,
        lb,
        ub,
        vtype: Vtype,
        name: str = "",
    ) -> IVar:
        """
        创建变量
        :param lb: 变量下界
        :param ub: 变量上界
        :param vtype: 变量类型： Vtype.CONTINUOUS（连续）,Vtype.BINARY(0-1变量), Vtype.BINARY（整数变量）
        :param name:变量名
        :return:变量实体
        """
        ...

    @abstractmethod
    def addVars(
        self,
        *indices,
        lb,
        ub,
        vtype: Vtype,
        name: str = "",
    ) -> ITupledict:
        """
        创建多维变量
        :param indices:多维的参数，如addVars(1,2),addVars(mList,nList),addVars([1,2,3],[3,4,5])等。
        :param lb:变量下界
        :param ub:变量上界
        :param vtype:变量类型： Vtype.CONTINUOUS（连续）,Vtype.BINARY(0-1变量), Vtype.BINARY（整数变量）
        :param name:变量名
        :return:tupledict类型
        """
        ...

    @abstractmethod
    def getVars(self) -> List[IVar]:
        """
        获取所有的变量对象
        :return:
        """
        ...

    @abstractmethod
    def getVar(self, i) -> IVar:
        """
        获取第i个变量
        :param i:
        :return:
        """
        ...

    @abstractmethod
    def addConstr(self, lin_expr):
        """
        向模型添加约束条件，
        :param lin_expr:线性约束表达式
        :param name: 约束名称
        :return:
        """
        ...

    @abstractmethod
    def addConstrs(self, lin_exprs: Union[List, Generator, Tuple]):
        """
        向模型添加多个约束条件，
        :param lin_exprs: 线性约束表达式集合 可以为列表或者元组。
        :param name:名称
        :return:
        """
        ...

    @abstractmethod
    def setObjective(self, expr):
        """
        设置模型的单一目标
        :param expr: 目标表达式
        优化方向。ObjType.MINIMIZE（最小值），ObjType.MAXIMIZE(最大值)
        :return:
        """
        self._multi_object_expr.clear()
        self._multi_object_expr.append((expr, 1, "单一目标"))
        if self._flag_objective_n is None:
            self._flag_objective_n = False
        else:
            if self._flag_objective_n:
                raise Exception("单一目标和多目标不能同时设置")
            else:
                raise Exception("单一目标设置多次,请检查")

    @abstractmethod
    def setObjectiveN(
        self, expr, index: int, priority: int = 0, weight: float = 1, name: str = ""
    ):
        """
        多目标优化，优化最小值
        :param expr: 表达式
        :param index: 目标函数对应的序号 (默认 0，1，2，…), 以 index=0 作为目标函数的值, 其余值需要另外设置参数
        :param priority: 分层序列法多目标决策优先级(整数值), 值越大优先级越高【未实现】
        :param weight: 线性加权多目标决策权重(在优先级相同时发挥作用)
        :param name: 名称
        :return:
        """
        if self._flag_objective_n is False:
            raise Exception("单一目标和多目标不能同时设置")
        else:
            self._flag_objective_n = True
        self._multi_object_expr.append((expr, weight, name))

    @abstractmethod
    def addGenConstrAnd(self, resvar, varList: List[IVar]):
        """
        and 运算。
        addGenConstrAnd(y, [x1, x2]) 表示y = max(x1,x2)。 所有变量均为0-1变量
        :param resvar:
        :param varList:
        :param name:
        :return:
        """
        check_bool_var(resvar, varList)
        for var in varList:
            self.addConstr(resvar <= var)
        self.addConstr(resvar >= self.Sum(varList) - len(varList) + 1)

    @abstractmethod
    def addGenConstrOr(self, resvar: IVar, varList: List[IVar]):
        """
        或运算
        addGenConstrOr(y, [x1, x2]) 表示y = min(x1,x2)。 所有变量均为0-1变量
        :param resvar:
        :param varList:
        :param name:
        :return:
        """
        check_bool_var(resvar, varList)
        for var in varList:
            self.addConstr(resvar >= var)
        self.addConstr(resvar <= self.Sum(varList))

    @abstractmethod
    def addGenConstrXOr(self, resvar: IVar, varList: List[IVar]):
        """
        异或运算
        addGenConstrXOr(y, [x1, x2])。 所有变量均为0-1变量
        :param resvar:
        :param varList:
        :param name:
        :return:
        """
        if len(varList) != 2:
            raise ValueError("length of vars must be 2")
        check_bool_var(resvar, varList)
        self.addConstr(resvar >= varList[0] - varList[1])
        self.addConstr(resvar >= varList[1] - varList[0])
        self.addConstr(resvar <= varList[0] + varList[1])
        self.addConstr(resvar <= 2 - varList[0] - varList[1])

    @abstractmethod
    def addGenConstrPWL(
        self, var_x: IVar, var_y: IVar, x_range: List[float], y_range: List, M
    ):
        """
        设置分段约束
        :param var_y: f(x)
        :param var_x:指定变量的目标函数是分段线性
        :param x:  定义分段线性目标函数的点的横坐标值(非减序列)
        :param y:定义分段线性目标函数的点的纵坐标值
        :return:
        """
        assert len(x_range) == len(y_range) - 1, "分段函数设置异常"
        # name = f"{var_x.VarName}_{var_y.VarName}"
        z = self.addVars(len(y_range), vtype=Vtype.BINARY, lb=0, ub=1)
        self.addConstr(z.quicksum() == 1)
        self.addConstr(
            var_x
            <= z.quickprod(
                {i: x_range[i] if i < len(z) - 1 else M for i in range(len(z))}, "*"
            )
        )
        self.addConstr(
            var_x
            >= z.quickprod(
                {i: x_range[i - 1] if i > 0 else -M for i in range(len(z))}, "*"
            )
        )

        for i in range(len(y_range)):
            self.addConstr(var_y <= y_range[i] + (1 - z[i]) * M)
            self.addConstr(var_y >= y_range[i] - (1 - z[i]) * M)

    @abstractmethod
    def addGenConstrIndicator(
        self,
        binvar: IVar,
        binval: bool,
        lhs: IVar,
        sense: CmpType,
        rhs,
        M,
    ):
        """
        若 binvar 为binval ,则 lhs 与 rhs 之间有sense 的关系
        若M不指定,则程序会给与默认。但仍推荐给出M。程序自动给出的可能会存在问题。
        :param binvar: 0-1变量
        :param binval: bool 常量
        :param lhs:  左侧变量
        :param sense: 等号，大于等于，小于等于
        :param rhs: 右侧常量
        :param M: 大M
        :return:
        """
        if binval is True:
            z = 1 - binvar
        else:
            z = binvar
        if sense == CmpType.GREATER_EQUAL:
            self.addConstr(lhs + M * z >= rhs)
        elif sense == CmpType.EQUAL:
            self.addConstr(lhs + M * z >= rhs)
            self.addConstr(lhs - M * z <= rhs)
        else:
            self.addConstr(lhs - M * z <= rhs)

    @abstractmethod
    def addIndicator(self, binvar: IVar, binval: bool, constr):
        """
        若 binvar 为binval ,则 lhs 与 rhs 之间有sense 的关系
        若M不指定，则程序会给与默认。但仍推荐给出M。程序自动给出的可能会存在问题。
        :param binvar: 0-1变量
        :param binval: bool 常量
        :param lhs:  左侧变量
        :param sense: 等号，大于等于，小于等于
        :param rhs: 右侧常量
        :param M: 大M
        :return:
        """
        ...

    @abstractmethod
    def addGenConstrAbs(self, resvar: any, var_abs: IVar, M):  # type: ignore
        """
        绝对值 resvar = |var_abs|
        :param resvar:
        :param var_abs:
        :return:
        """
        ...

    @abstractmethod
    def addGenConstrMultiply(self, z: IVar, l: Tuple[IVar, IVar]):
        """
        满足 z = x * y
        其中 x 为0,1变量
        :param l:
        :param z: 变量
        :return:
        """
        if len(list(l)) != 2:
            raise RuntimeError("Only need two variables")

    @abstractmethod
    def addRange(self, expr, min_value: Real, max_value: Real):
        """
        添加范围约束
        :param expr: 表达式
        :param min_value: 最小值
        :param max_value: 最大值
        :return:
        """
        if isinstance(min_value, (int, float)) and isinstance(max_value, (int, float)):
            if min_value > max_value:
                raise ValueError("min_value is bigger than max_value")
        self.addConstr(expr >= min_value)
        self.addConstr(expr <= max_value)

    @abstractmethod
    def addConstrOr(
        self,
        constrs: List,
        ok_num: int = 1,
        cmp_type: CmpType = CmpType.EQUAL,
        M=None,
    ):
        """
        约束的满足情况,满足的次数
        :param constr: 所有的约束
        :param ok_num:  需要满足的个数，具体则根据cmpType
        :param cmpType: CmpType.LESS_EQUAL CmpType.EQUAL,CmpType.GREATER_EQUAL
        :param M: M值，推荐指定M值。
        :return:
        """
        ...

    @abstractmethod
    def numVars(self, vtype: Optional[Vtype] = None) -> int:
        """
        变量个数
        :return:
        """
        ...

    @abstractmethod
    def numConstraints(self) -> int:
        """
        约束个数
        :return:
        """
        ...

    @abstractmethod
    def write(self, filename: str, obfuscated=False):
        """
        写入到文件
        :param filename:文件名，支持后缀 .lp .mps .proto(目前有问题)
        :param obfuscated: 是否混淆，默认不混淆
        :return:
        """
        ...

    @abstractmethod
    def read(self, path: str):
        """读取文件 lp,mps 文件。

        Args:
            path (str): _description_
        """
        ...

    @property
    @abstractmethod
    def ObjVal(self) -> Real:
        """目标值"""
        ...

    @abstractmethod
    def optimize(self, obj_type: ObjType = ObjType.MINIMIZE) -> OptimizationStatus:
        """
        优化目标
        :param time_limit_milliseconds:最大运行时长
        :param obj_type:优化目标。ObjType.MINIMIZE（最小值），ObjType.MAXIMIZE(最大值)
        :param enable_output: 是否显示gap日志。
        :return:
        """
        ...

    @abstractmethod
    def clear(self):
        """清空所有的目标和约束"""
        self._multi_object_expr.clear()

    @abstractmethod
    def close(self):
        """关闭"""
        ...

    @abstractmethod
    def valueExpression(self, expression) -> Real:
        """
        计算表达式的值。
        :param expression:
        :return:
        """
        ...

    @abstractmethod
    def newIntervalVar(self, start, size, end) -> Any:
        """
        创建变量： start+size=end

        Args:
            start (_type_): 开始
            size (_type_): 大小
            end (_type_): 结束
        """
        ...

    @abstractmethod
    def addNoOverlap(self, interval_vars: List, M):
        """
        互相之间不重复

        Args:
            interval_vars (List): 间隔变量
        """
        pass

    @abstractmethod
    def setNumThreads(self, num_theads: int):
        """
        设置线程的个数

        Args:
            num_theads (int): 线程个数
        """
        ...


class ICpModel(IModel):
    def __init__(self, solver_id: Literal["CPLEX_CP", "CP"], name=""):
        super().__init__(
            solver_id,
            name,
        )

    @abstractmethod
    def addCpElement(self, index, variables, target): ...

    @abstractmethod
    def addCpCircuit(self, arcs): ...

    @abstractmethod
    def addCpAllowedAssignments(self, variables, tuples_list): ...

    @abstractmethod
    def addCpForbiddenAssignments(self, variables, tuples_list): ...

    @abstractmethod
    def addCpInverse(
        self,
        variables,
        inverse_variables,
    ): ...

    @abstractmethod
    def addCpMapDomain(self, var, bool_var_array, offset):
        """Adds `var == i + offset <=> bool_var_array[i] == true for all i`."""
        ...

    @abstractmethod
    def addCpImplication(self, a, b): ...

    @abstractmethod
    def addCpBoolTrueOr(self, literals): ...

    @abstractmethod
    def addCpBoolTrueAnd(self, literals): ...

    @abstractmethod
    def addCpAtLeastOneIsTrue(self, literals): ...

    @abstractmethod
    def addCpAtMostOneIsTrue(self, literals): ...

    @abstractmethod
    def addCpExactlyNumIsTrue(self, literals, num: int = 1): ...

    @abstractmethod
    def addCpGenConstrMax(self, target: IVar, varList: List[IVar]):
        """
        最大值约束
        """
        ...

    @abstractmethod
    def addCpGenConstrMin(self, target: IVar, varList: List[IVar]):
        """
        最小值约束
        """
        ...

    @abstractmethod
    def addCpGenConstrDivision(self, target, num, denom):
        """Adds `target == num // denom` (integer division rounded towards 0)."""
        ...

    @abstractmethod
    def addCpGenConstrModulo(self, target, num, denom):
        """Adds `target = expr % mod`."""
        ...

    @abstractmethod
    def addCpGenConstrMultiplication(
        self,
        target,
        *expressions,
    ): ...



# -*- coding: utf-8 -*-
# @Time    : 2023/4/1 10:41
# @Author  : luyi
from collections import UserList
from typing import List, Iterable
from warnings import warn

from .interface_ import ITuplelist
from .utilx import get_length, is_list_or_tuple


class tuplelist(ITuplelist):
    """
    For example:
      > l = tuplelist([(1, 2), (1, 3), (2, 3), (2, 4)])
      > print(l.select('*', '*'))
      [(1, 2), (1, 3), (2, 3), (2, 4)]
      > print(l.select('*', 3))
      [(1, 3), (2, 3)]
      > print(l.select(1, '*'))
      [(1, 2), (1, 3)]
    """

    def append(self, item):
        # 检查是否重复。若重复则不添加。
        if isinstance(item, UserList) or type(item) == type(
            []
        ):  # 由于特殊性，不进行过多的扩展。
            raise ValueError("append传入参数不能为集合")
        if item in self.data:
            return
        # 检查维度
        if self.dim is None:
            self.dim = get_length(item)
        else:
            if self.dim != get_length(item):
                raise ValueError("传入的数据维度不一致。")
        super().append(item)

    def extend(self, other: Iterable[any]):  # type: ignore
        # 检查是否重复。
        if isinstance(other, UserList):
            datas = other.data
        else:
            datas = other
        for item in datas:
            self.append(item)

    def __init__(self, seq=None):
        seq = self._duplicate(seq)
        if len(seq) == 0:
            self.dim = None
            super().__init__([])
        else:
            self.dim = get_length(seq[0])
            super().__init__(seq)

    def _get_dim(self):
        if self.dim is None:
            if len(self) == 0:
                self.dim = 0
                return self.dim
            self.dim = get_length(self[0])
        return self.dim

    def quickselect(self, *pattern) -> List:
        dim = len(pattern)
        if dim == 0:
            return list(self)  # type: ignore
        if len(self) == 0:
            return []
        li = []
        if self._get_dim() == 1:
            for key in self:
                p = pattern[0]
                if p == "*" or p == key:
                    li.append(key)
            return li
        for key in self:
            ok = True
            # 判定key是否可以迭代
            for i, k in enumerate(key):
                p = pattern[i]
                if p == "*":
                    continue
                if p != k:
                    ok = False
                    break
            if ok:
                li.append(key)
        return li

    def select(self, *args) -> List:
        warn(f"call to deprecated function,use quickselect function")
        values = [*args]
        if len(self.data) == 0:
            return []
        if len(values) == 0:  # 全选
            return self.data
        if self.dim is None:
            self.dim = len(values)
        if self.dim != len(values):
            raise ValueError("传入的数据个数异常")
        _dim_ok_index = []
        _list_num_range = [i for i in range(len(self.data))]
        # 为了减少其他方法如pop,remove 等操作。将计算的操作放入该位置。
        # 会存在重复计算的问题。
        _keys = []
        for i in range(self.dim):
            temp = []
            for tu in self.data:
                _dim = get_length(tu)
                # 检查传入的参数维度是否一致
                if _dim != self.dim:
                    raise ValueError("传入的数据维度不一致。")
                if is_list_or_tuple(tu):
                    temp.append(tu[i])
                else:
                    temp.append(tu)
            _keys.append(temp)

        for i, key in enumerate(values):
            if key == "*":
                _dim_ok_index.append(_list_num_range)
                continue
            if isinstance(key, (float, int, str)):
                key = [key]
            _dim_ok_index.append([j for j, x in enumerate(_keys[i]) if x in key])
        flag = []
        for i, item in enumerate(_dim_ok_index):
            if i == 0:
                flag = item
                continue
            flag = list(set(flag).intersection(set(item)))
        return [key for i, key in enumerate(self.data) if i in flag]

    @staticmethod
    def _duplicate(initList):
        data = []
        if initList is not None:
            if type(initList) == type(data):
                data[:] = initList
            elif isinstance(initList, UserList):
                data[:] = initList.data[:]
            else:
                data = list(initList)
        l = [key for key in data]
        dim = None  # 检查传入的数据是否都是同一纬度的。
        for i, key in enumerate(l):
            if isinstance(key, list):
                t_key = tuple(key)
                l[i] = t_key
            _dim = get_length(key)
            if dim is None:
                dim = _dim
            else:
                if dim != _dim:
                    raise ValueError("传入的数据维度不一致。")
        return list(set(l))


if __name__ == "__main__":
    a = tuplelist([(1, 2), (1, 2), [2, 3]])
    print(a.quickselect())
    print({1, 1})



import warnings
import gurobipy as gp
from typing import Any, Dict, Generator, List, Optional, Tuple, Union

from numpy import isin
from mipx.constants import CmpType, ObjType, OptimizationStatus, Vtype
from mipx.interface_ import IModel, IVar
from typing_extensions import override

from mipx.tupledict import tupledict

from mipx.utilx import INFINITY, is_bool_var


Real = Union[float, int]

gp.tupledict.quicksum = gp.tupledict.sum  # type: ignore
gp.tupledict.quickprod = gp.tupledict.prod  # type: ignore
gp.tupledict.quickselect = gp.tupledict.select  # type: ignore


class GurobiSolver(IModel):
    def __init__(self, name=""):
        super().__init__("CPLEX", name)
        self.__model = gp.Model(name)

    @override
    def Sum(self, expr_array):
        return gp.quicksum(expr_array)

    @override
    def setHint(self, start: Dict[IVar, Real]):
        pass

    @override
    def setTimeLimit(self, time_limit_seconds: int):
        self.__model.Params.TimeLimit = time_limit_seconds

    @override
    def wall_time(self) -> float:
        return self.__model.Runtime

    @override
    def iterations(self) -> Real:
        raise NotImplementedError()

    @override
    def nodes(self) -> Real:
        raise NotImplementedError()

    @override
    def addVar(
        self,
        lb: Real = 0.0,
        ub: Real = INFINITY,
        vtype: Vtype = Vtype.CONTINUOUS,
        name: str = "",
    ) -> IVar:
        """
            Add a decision variable to a model.
        :param lb: Lower bound for new variable.
        :param ub: Upper bound for new variable.
        :param vtype: variable type for new variable(Vtype.CONTINUOUS, Vtype.BINARY, Vtype.INTEGER).
        :param name: Name for new variable.
        :return: variable.
        """
        gvtype = gp.GRB.CONTINUOUS
        if vtype == Vtype.BINARY:
            gvtype = gp.GRB.BINARY
        elif vtype == Vtype.INTEGER:
            gvtype = gp.GRB.INTEGER
        var = self.__model.addVar(lb=lb, ub=ub, vtype=gvtype, name=name)
        return var  # type: ignore

    @override
    def addVars(
        self,
        *indices,
        lb: Real = 0.0,
        ub: Real = INFINITY,
        vtype: Vtype = Vtype.CONTINUOUS,
        name: str = "",
    ) -> tupledict:
        gvtype = gp.GRB.CONTINUOUS
        if vtype == Vtype.BINARY:
            gvtype = gp.GRB.BINARY
        elif vtype == Vtype.INTEGER:
            gvtype = gp.GRB.INTEGER
        tu_dict = self.__model.addVars(*indices, lb=lb, ub=ub, name=name, vtype=gvtype)
        return tu_dict  # type: ignore

    @override
    def getVars(self) -> List[IVar]:
        return self.__model.getVars()  # type: ignore

    @override
    def getVar(self, i) -> IVar:  # type: ignore
        self.__model.getVarByName(i)

    @override
    def addConstr(self, lin_expr):
        self.__model.addConstr(lin_expr)

    @override
    def addConstrs(self, lin_exprs: Union[List, Generator, Tuple]):
        if isinstance(lin_exprs, Generator):
            self.__model.addConstrs(lin_exprs)
        else:
            self.__model.addConstrs(constr for constr in lin_exprs)

    @override
    def setObjective(self, expr):
        self.__model.setObjective(expr)

    @override
    def setObjectiveN(
        self, expr, index: int, priority=None, weight: float = 1, name: str = ""
    ):
        """

        :param expr:
        :param index:
        :param priority:
        :param weight:
        :param name:
        :return:
        """
        self.__model.setObjectiveN(expr, index, priority, weight, name=name)

    @override
    def addGenConstrAnd(self, resvar, varList: List[IVar]):
        """
        和 addGenConstrAnd(y, [x1,x2])
        :param resvar:
        :param varList:
        :return:
        """
        self.__model.addGenConstrAnd(resvar, varList)  # type: ignore

    @override
    def addGenConstrOr(self, resvar: IVar, varList: List[IVar]):
        self.__model.addGenConstrOr(resvar, varList)  # type: ignore

    @override
    def addGenConstrXOr(self, resvar: IVar, varList: List[IVar]):
        super().addGenConstrXOr(resvar, varList)

    @override
    def addGenConstrPWL(
        self, var_x: IVar, var_y: IVar, x_range: List[float], y_range: List, M
    ):
        super().addGenConstrPWL(var_x, var_y, x_range, y_range, M)

    @override
    def addGenConstrIndicator(
        self, binvar: IVar, binval: bool, lhs: IVar, sense: CmpType, rhs: float, M=None
    ):
        """
        若 binvar 为binval ,则 lhs 与 rhs 之间有sense 的关系

        :param binvar: 0-1变量
        :param binval: bool 常量
        :param lhs:  左侧变量
        :param sense: 等号，大于等于，小于等于
        :param rhs: 右侧常量
        :param M: 大M
        :return:
        """
        g_sense = gp.GRB.LESS_EQUAL
        if sense == CmpType.GREATER_EQUAL:
            g_sense = gp.GRB.GREATER_EQUAL
        elif sense == CmpType.LESS_EQUAL:
            g_sense = gp.GRB.LESS_EQUAL
        self.__model.addGenConstrIndicator(binvar, binval, lhs, g_sense, rhs)  # type: ignore

    @override
    def addIndicator(self, binvar: IVar, binval: bool, constr):
        raise NotImplementedError()

    @override
    def addGenConstrAbs(self, resvar, var_abs: IVar, M):
        if isinstance(var_abs, gp.Var):
            self.__model.addGenConstrAbs(resvar, var_abs)  # type: ignore
        else:
            temp = self.addVar()
            self.addConstr(temp == var_abs)
            self.addGenConstrAbs(resvar, temp, M)

    @override
    def addGenConstrMultiply(self, z: IVar, l: Tuple[IVar, IVar]):
        """x * y = z"""
        warnings.warn("gurobi使用下该方法<addConstrMultiply>可能会影响性能")
        self.__model.update()
        super().addGenConstrMultiply(z, l)
        x = l[0]
        y = l[1]
        if not is_bool_var(x) and not is_bool_var(y):
            raise RuntimeError("At least one binary variable is required.")
        if is_bool_var(y):
            x, y = y, x
        M = y.UB
        self.addConstr(z <= y)
        self.addConstr(z <= x * M)
        self.addConstr(z >= y + (x - 1) * M)

    @override
    def addRange(
        self, expr, min_value: Union[float, int], max_value: Union[float, int]
    ):
        self.__model.addRange(expr, min_value, max_value)

    @override
    def calc_lbub_and_expr(self, constr) -> Tuple[Real, Real, Any]:
        raise NotImplementedError()

    @override
    def addConstrOr(
        self,
        constrs: List,
        ok_num: int = 1,
        cmp_type: CmpType = CmpType.EQUAL,
        M=None,
    ):
        """
        约束的满足情况

        :param constr: 所有的约束
        :param ok_num: 需要满足的个数，具体则根据cmpType
        :param cmpType: CmpType.LESS_EQUAL CmpType.EQUAL,CmpType.GREATER_EQUAL
        :return:
        """
        constr_num = len(constrs)
        x = []
        for i in range(constr_num):
            x.append(self.addVar(vtype=Vtype.BINARY))
            constr = constrs[i]
            lb, ub, expr = self.calc_lbub_and_expr(constr)
            tempM = self.INFINITY
            if M is not None:
                tempM = M
            if lb > -self.INFINITY:  # 若大于
                self.addConstr(expr + tempM * (1 - x[i]) >= lb)
            if ub < self.INFINITY:
                self.addConstr(expr - tempM * (1 - x[i]) <= ub)

        if cmp_type == CmpType.EQUAL:
            self.addConstr(self.Sum(x) == ok_num)
        elif cmp_type == CmpType.GREATER_EQUAL:
            self.addConstr(self.Sum(x) >= ok_num)
        elif cmp_type == CmpType.LESS_EQUAL:
            self.addConstr(self.Sum(x) <= ok_num)  # type: ignore
        else:
            raise Exception("error value of cmpType")

    @override
    def numVars(self, vtype: Optional[Vtype] = None) -> int:
        if vtype is None:
            return self.__model.NumVars
        elif vtype == Vtype.CONTINUOUS:
            return self.__model.NumVars - self.__model.NumIntVars
        elif vtype == Vtype.INTEGER:
            return self.__model.NumIntVars - self.__model.NumBinVars
        else:
            return self.__model.NumBinVars

    @override
    def numConstraints(self) -> int:
        return self.__model.NumConstrs

    @override
    def write(self, filename: str, obfuscated=False):
        self.__model.write(filename)

    @override
    def read(self, path: str):
        self.__model.read(path)

    @property
    @override
    def ObjVal(self) -> Real:
        return self.__model.ObjVal

    @override
    def optimize(self, obj_type: ObjType = ObjType.MINIMIZE) -> OptimizationStatus:
        if self._params.TimeLimit:
            self.setTimeLimit(self._params.TimeLimit)
        if self._params.MIPGap:
            self.__model.Params.MIPGap = self._params.MIPGap
        if self._params.Precision:
            pass
        if self._params.EnableOutput:
            self.__model.setParam("OutputFlag", 1)
        else:
            self.__model.setParam("OutputFlag", 0)
        self.__model.ModelSense = (
            gp.GRB.MAXIMIZE if obj_type == ObjType.MAXIMIZE else gp.GRB.MINIMIZE
        )
        self.__model.optimize()
        status = self.__model.Status
        if status == gp.GRB.OPTIMAL:
            result = OptimizationStatus.OPTIMAL
        elif status == gp.GRB.INFEASIBLE:
            result = OptimizationStatus.INFEASIBLE
        elif status == gp.GRB.UNBOUNDED:
            result = OptimizationStatus.UNBOUNDED
        elif status == gp.GRB.CUTOFF:
            result = OptimizationStatus.CUTOFF
        elif status == gp.GRB.LOADED:
            result = OptimizationStatus.LOADED
        elif status == gp.GRB.TIME_LIMIT:
            result = OptimizationStatus.FEASIBLE
        else:
            result = OptimizationStatus.ERROR
        return result

    @override
    def clear(self):
        self.__model.clear()

    @override
    def close(self):
        self.__model.end()

    # 独有。

    @override
    def addNoOverlap(self, interval_vars, M):
        """
        相邻不重复。

        :param interval_vars:
        :return:
        """
        raise RuntimeError("not implemented")

    @override
    def newIntervalVar(self, start, size, end):
        raise RuntimeError("not implemented")

    @override
    def valueExpression(self, expression):

        return expression.getValue()

    @override
    def setNumThreads(self, num_theads: int):
        self.__model.setParam("Threads", num_theads)



# -*- coding: utf-8 -*-
# @Time    : 2023/4/9 18:21
# @Author  : luyi
from setuptools import setup

# with open("README.md", "r", encoding="utf-8") as f:
#     long_description = f.read()
setup(
    name="mipx",
    version="0.6.3",
    author="ly",
    author_email="2662017230@qq.com",
    # description="Python (Mixed-Integer Linear Programming Constraint Programming) Optimization Tools",
    description="",
    url="",
    # 你要安装的包，通过 setuptools.find_packages 找到当前目录下有哪些包
    # packages=find_packages(exclude=['core', '__pycache__']),
    packages=["mipx"],
    # long_description=long_description,
    # long_description_content_type="text/markdown",  # 详细描述的内容类型
    # include_package_data=True,
    install_requires=[
        "ortools>=9.9.3963,<=9.14.6206",
    ],
    extras_require={
        "docplex": ["docplex>=2.25.0,<3.0.0"],  # 限制 docplex 的版本范围
    },
    python_requires=">=3.8",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: Apache Software License",
    ],
    zip_safe=False,
)



from typing import (
    Any,
    Dict,
    Generator,
    Iterable,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
)
from typing_extensions import override
from .utilx import pre_condition, get_update_lub
from .interface_ import ICpModel, ITupledict, IVar
from .cpsolver import CpModelSolver
from .constants import CmpType, ObjType, OptimizationStatus, Vtype, Params

Real = Union[float, int]

# Type aliases
IntegralT = int
IntegralTypes = (int,)
NumberT = Union[
    int,
    float,
]
NumberTypes = (
    int,
    float,
)

LiteralT = Union["IVar", IntegralT, bool]
BoolVarT = IVar
VariableT = Union["IVar", IntegralT]

ArcT = Tuple[IntegralT, IntegralT, LiteralT]


class CpModel(ICpModel):
    """模型接口"""

    def __init__(self, solver_id: Literal["CP", "CPLEX_CP"] = "CP", name=""):
        super().__init__(solver_id)
        self.solver_id = solver_id
        if solver_id == "CP":
            self.__model = CpModelSolver(name=name)
        elif solver_id == "CPLEX_CP":
            try:
                from .cplexcpsolver import CplexCpSolver
            except ImportError:
                raise ImportError(
                    "Docplex not detected, please install docplex and rerun the program."
                )

            self.__model = CplexCpSolver(name=name)

    @property
    def Params(self) -> Params:
        """
        获取参数设置
        :return:
        """
        return self.__model.Params

    @override
    def Sum(self, expr_array) -> Any:
        if isinstance(expr_array, Generator):
            expr_array = list(expr_array)
        return self.__model.Sum(expr_array)

    @override
    def setHint(self, start: Dict[IVar, Real]):
        self.__model.setHint(start)  # type: ignore

    @override
    def setTimeLimit(self, time_limit_seconds: int):
        """
        设置程序最大运行时间
        :param time_limit_seconds: 秒
        """
        self.__model.setTimeLimit(time_limit_seconds)

    @override
    def wall_time(self) -> float:
        """
        求解所花费的时间
        :return: 求解所花费的时间，单位毫秒
        """
        return self.__model.wall_time()

    @override
    def iterations(self):
        """

        :return: 算法迭代的次数
        """
        return self.__model.iterations()

    @override
    def nodes(self) -> Real:
        """

        :return: 节点数
        """
        return self.__model.nodes()

    @override
    def addVar(
        self,
        lb=0,
        ub=None,
        vtype: Vtype = Vtype.INTEGER,
        name: str = "",
    ) -> IVar:
        """
        创建变量
        :param lb: 变量下界
        :param ub: 变量上界
        :param vtype: 变量类型： Vtype.CONTINUOUS（连续）,Vtype.BINARY(0-1变量), Vtype.BINARY（整数变量）
        :param name:变量名
        :return:变量实体
        """
        lb, ub = get_update_lub(True, lb, ub, vtype)
        if vtype == Vtype.INTEGER:
            lb = int(lb + 0.5)
            ub = int(ub)
        if lb > ub:
            raise ValueError("变量的上下界的取值不正确")
        return self.__model.addVar(lb=lb, ub=ub, vtype=vtype, name=name)

    @override
    def addVars(
        self,
        *indices,
        lb: int = 0,
        ub=None,
        vtype: Vtype = Vtype.INTEGER,
        name: str = "",
    ) -> ITupledict:
        """
        创建多维变量
        :param indices:多维的参数，如addVars(1,2),addVars(mList,nList),addVars([1,2,3],[3,4,5])等。
        :param lb:变量下界
        :param ub:变量上界
        :param vtype:变量类型： Vtype.CONTINUOUS（连续）,Vtype.BINARY(0-1变量), Vtype.BINARY（整数变量）
        :param name:变量名
        :return:tupledict类型
        """
        pre_condition(len(indices) > 0, "addVars中多维参数缺失")
        lb, ub = get_update_lub(True, lb, ub, vtype)
        return self.__model.addVars(*indices, lb=lb, ub=ub, vtype=vtype, name=name)

    @override
    def getVars(self) -> List[IVar]:
        """
        获取所有的变量对象
        :return:
        """
        return self.__model.getVars()

    @override
    def getVar(self, i) -> IVar:
        """
        获取第i个变量
        :param i:
        :return:
        """
        return self.__model.getVar(i)

    @override
    def addConstr(self, lin_expr):
        """
        向模型添加约束条件，
        :param lin_expr:线性约束表达式
        :return:
        """
        return self.__model.addConstr(lin_expr)

    @override
    def addConstrs(self, lin_exprs: Union[List, Generator, Tuple]):
        """
        向模型添加多个约束条件，
        :param lin_exprs: 线性约束表达式集合 可以为列表或者元组。
        :return:
        """
        return self.__model.addConstrs(lin_exprs)

    @override
    def setObjective(self, expr):
        """
        设置模型的单一目标
        :param expr: 目标表达式
        优化方向。ObjType.MINIMIZE（最小值），ObjType.MAXIMIZE(最大值)
        :return:
        """
        super().setObjective(expr)
        self.__model.setObjective(expr)

    @override
    def setObjectiveN(
        self, expr, index: int, priority: int = 0, weight: float = 1, name: str = ""
    ):
        """
        多目标优化，优化最小值
        :param expr: 表达式
        :param index: 目标函数对应的序号 (默认 0，1，2，…), 以 index=0 作为目标函数的值, 其余值需要另外设置参数
        :param priority: 分层序列法多目标决策优先级(整数值), 值越大优先级越高【未实现】
        :param weight: 线性加权多目标决策权重(在优先级相同时发挥作用)
        :param name: 名称
        :return:
        """
        super().setObjectiveN(expr, index, priority, weight, name)
        self.__model.setObjectiveN(
            expr, index, priority=priority, weight=weight, name=name
        )

    @override
    def addGenConstrAnd(self, resvar, varList: List[IVar]):
        """
        and 运算。
        addGenConstrAnd(y, [x1, x2]) 表示y = min(x1,x2)。 所有变量均为0-1变量
        :param resvar:
        :param varList:
        :return:
        """
        self.__model.addGenConstrAnd(resvar, varList)

    @override
    def addGenConstrOr(self, resvar: IVar, varList: List[IVar]):
        """
        或运算
        addGenConstrOr(y, [x1, x2]) 表示y = max(x1,x2)。 所有变量均为0-1变量
        :param resvar:
        :param varList:
        :return:
        """
        self.__model.addGenConstrOr(resvar, varList)

    @override
    def addGenConstrXOr(self, resvar: IVar, varList: List[IVar]):
        """
        异或运算
        addGenConstrXOr(y, [x1, x2])。 所有变量均为0-1变量
        x1 x2 y
        1  1  0
        1  0  1
        0  1  1
        0  0  0
        :param resvar:
        :param varList:
        :return:
        """
        self.__model.addGenConstrXOr(resvar, varList)

    @override
    def addGenConstrPWL(
        self, var_x: IVar, var_y: IVar, x_range: List[float], y_range: List, M=None
    ):
        """
        设置分段约束
        :param var_y: f(x)
        :param var_x:指定变量的目标函数是分段线性
        :param x:  定义分段线性变量的点的范围边界(非减序列)，要求x的范围大于0
        :param y:定义分段线性变量的范围所对应的目标函数的值
        :return:
        """
        if M is None:
            M = self.__model.INFINITY
        self.__model.addGenConstrPWL(var_x, var_y, x_range, y_range, M)

    @override
    def addGenConstrIndicator(
        self,
        binvar: IVar,
        binval: bool,
        lhs: IVar,
        sense: CmpType,
        rhs,
        M,
    ):
        """
        若 binvar 为binval ,则 lhs 与 rhs 之间有sense 的关系
        若M不指定，则程序会给与默认。但仍推荐给出M。程序自动给出的可能会存在问题。
        :param binvar: 0-1变量
        :param binval: bool 常量
        :param lhs:  左侧变量
        :param sense: 等号，大于等于，小于等于
        :param rhs: 右侧常量
        :param M: 大M
        :return:
        """
        self.__model.addGenConstrIndicator(binvar, binval, lhs, sense, rhs, M=M)

    @override
    def addIndicator(self, binvar: IVar, binval: bool, constr):
        self.__model.addIndicator(binvar, binval, constr)

    @override
    def addGenConstrAbs(self, resvar, var_abs: IVar, M=None):
        """
        绝对值 resvar = |var_abs|
        :param resvar:
        :param var_abs:
        :return:
        """
        self.__model.addGenConstrAbs(resvar, var_abs)

    @override
    def addGenConstrMultiply(self, z: IVar, l: Tuple[IVar, IVar]):
        """
        满足 z = x * y
        其中 x 为0,1变量
        :param l:
        :param z: 变量
        :return:
        """
        self.__model.addGenConstrMultiply(z, l)

    @override
    def addRange(
        self, expr, min_value: Union[float, int], max_value: Union[float, int]
    ):
        """
        添加范围约束
        :param expr: 表达式
        :param min_value: 最小值
        :param max_value: 最大值
        :return:
        """
        self.__model.addRange(expr, min_value, max_value)

    @override
    def addConstrOr(
        self,
        constrs: List,
        ok_num: int = 1,
        cmp_type: CmpType = CmpType.EQUAL,
        M=None,
    ):
        """
        约束的满足情况,满足的次数
        :param constr: 所有的约束
        :param ok_num:  需要满足的个数，具体则根据cmpType
        :param cmpType: CmpType.LESS_EQUAL CmpType.EQUAL,CmpType.GREATER_EQUAL
        :param M: M值，推荐指定M值。
        :return:
        """
        self.__model.addConstrOr(constrs, ok_num=ok_num, cmp_type=cmp_type, M=M)

    @override
    def numVars(self, vtype: Optional[Vtype] = None) -> int:
        """
        变量个数
        :return:
        """
        return self.__model.numVars(vtype)

    @override
    def numConstraints(self) -> int:
        """
        约束个数
        :return:
        """
        return self.__model.numConstraints()

    @override
    def write(self, filename: str, obfuscated=False):
        """
        写入到文件
        :param filename:文件名，支持后缀 .lp .mps .proto(目前有问题)
        :param obfuscated: 是否混淆，默认不混淆
        :return:
        """
        self.__model.write(filename, obfuscated=obfuscated)

    @override
    def read(self, path: str):
        self.__model.read(path)

    @property
    def ObjVal(self) -> Real:
        """目标值"""
        return self.__model.ObjVal

    @override
    def optimize(self, obj_type: ObjType = ObjType.MINIMIZE) -> OptimizationStatus:
        """
        优化目标
        :param time_limit_milliseconds:最大运行时长
        :param obj_type:优化目标。ObjType.MINIMIZE（最小值），ObjType.MAXIMIZE(最大值)
        :param enable_output: 是否显示gap日志。
        :return:
        """
        return self.__model.optimize(obj_type=obj_type)

    @override
    def calc_lbub_and_expr(self, constr) -> Tuple[Real, Real, Any]:
        return self.__model.calc_lbub_and_expr(constr)

    @override
    def clear(self):
        self.clearKpis()
        self.__model.clear()

    @override
    def close(self):
        self.__model.close()

    @override
    def valueExpression(self, expression):
        """
        计算表达式的值。
        :param expression:
        :return:
        """
        if isinstance(expression, (int, float)):
            return expression
        return self.__model.valueExpression(expression)

    @override
    def newIntervalVar(self, start, size, end):
        """
        创建变量： start+size=end

        Args:
            start (_type_): 开始
            size (_type_): 大小
            end (_type_): 结束
            name (str, optional): 名称. Defaults to "".
        """
        return self.__model.newIntervalVar(
            start,
            size,
            end,
        )

    @override
    def addNoOverlap(self, interval_vars: List, M: int):
        """
        互相之间不重复

        Args:
            interval_vars (List): 间隔变量
        """
        self.__model.addNoOverlap(interval_vars, M)

    @override
    def setNumThreads(self, num_theads: int):
        """
        设置线程的个数

        Args:
            num_theads (int): 线程个数
        """
        self.__model.setNumThreads(num_theads)

    @override
    def addCpGenConstrMin(self, target: IVar, varList: List[IVar]):
        self.__model.addCpGenConstrMin(target, varList)

    @override
    def addCpGenConstrMax(self, target: IVar, varList: List[IVar]):
        self.__model.addCpGenConstrMax(target, varList)

    @override
    def addCpGenConstrMultiplication(
        self,
        target,
        *expressions,
    ):
        """Adds `target == expressions[0] * .. * expressions[n]`."""
        self.__model.addCpGenConstrMultiplication(target, *expressions)

    @override
    def addCpGenConstrDivision(self, target, num, denom):
        """Adds `target == num // denom` (integer division rounded towards 0)."""
        self.__model.addCpGenConstrDivision(target, num, denom)

    @override
    def addCpGenConstrModulo(self, target, num, denom):
        """Adds `target = expr % mod`."""
        self.__model.addCpGenConstrModulo(target, num, denom)

    @override
    def addCpElement(
        self,
        index: Union[IVar, int],
        variables: Sequence[Union[IVar, int]],
        target: Union[IVar, int],
    ):
        """
        variables[index] = target

        :param Union[IVar, int] index: 指针
        :param Sequence[Union[IVar, int]] variables: 变量集合
        :param Union[IVar, int] target: 目标值
        """
        self.__model.addCpElement(index, variables, target)

    @override
    def addCpCircuit(self, arcs: Sequence[ArcT]):
        self.__model.addCpCircuit(arcs)

    @override
    def addCpAllowedAssignments(
        self,
        variables: Sequence[VariableT],
        tuples_list: Iterable[Sequence[IntegralT]],
    ):
        """
        限制允许的赋值
        要求：variables=[x1,x2,x3] 可选的值为tuples_list=[[1,2,3],[4,5,6],[7,8,9]]
        :param Sequence[VariableT] variables: 待赋值的变量集合
        :param Iterable[Sequence[IntegralT]] tuples_list: 可选的赋值集合
        """
        self.__model.addCpAllowedAssignments(variables, tuples_list)

    @override
    def addCpForbiddenAssignments(
        self,
        variables: Sequence[VariableT],
        tuples_list: Iterable[Sequence[IntegralT]],
    ):
        """
        禁止的赋值
        要求：variables=[x1,x2,x3] 不可选的值为tuples_list=[[1,2,3],[4,5,6],[7,8,9]]
        :param Sequence[VariableT] variables: 待赋值的变量集合
        :param Iterable[Sequence[IntegralT]] tuples_list: 不可选的赋值集合
        """
        self.__model.addCpForbiddenAssignments(variables, tuples_list)

    @override
    def addCpInverse(
        self,
        variables: Sequence[VariableT],
        inverse_variables: Sequence[VariableT],
    ):
        """
        可逆约束
        x[i]=j and x[j]=i
        """
        self.__model.addCpInverse(variables, inverse_variables)

    @override
    def addCpMapDomain(
        self, var: IVar, bool_var_array: Iterable[IVar], offset: IntegralT = 0
    ):
        """Adds `var == i + offset <=> bool_var_array[i] == true for all i`."""
        self.__model.addCpMapDomain(var, bool_var_array, offset)

    @override
    def addCpImplication(self, a: LiteralT, b: LiteralT):
        """
        Adds `a => b` (`a` implies `b`).
        即：若a为真，则b一定为真。
        """
        self.__model.addCpImplication(a, b)

    @override
    def addCpBoolTrueOr(self, literals: Iterable[LiteralT]):
        """Adds `Or(literals) == true`: sum(literals) >= 1."""
        self.__model.addCpBoolTrueOr(literals)

    @override
    def addCpBoolTrueAnd(self, literals: Iterable[LiteralT]):
        """全部都是真"""
        self.__model.addCpBoolTrueAnd(literals)

    @override
    def addCpAtLeastOneIsTrue(self, literals: Iterable[LiteralT]):
        self.__model.addCpAtLeastOneIsTrue(literals)

    @override
    def addCpAtMostOneIsTrue(self, literals: Iterable[LiteralT]):
        self.__model.addCpAtMostOneIsTrue(literals)

    @override
    def addCpExactlyNumIsTrue(self, literals: Iterable[LiteralT], num: int = 1):
        self.__model.addCpExactlyNumIsTrue(literals, num)



# -*- coding: utf-8 -*-
# @Time    : 2023/4/7 8:07
# @Author  : luyi
"""
变量
"""
from docplex.mp.dvar import Var as DVar
from docplex.cp.model import CpoIntVar
from docplex.cp.expression import _build_int_var_domain
from docplex.mp.kpi import DecisionKPI
from mipx.interface_ import IVar


class CPlexVar(DVar, IVar):

    # def Not(self) -> "IVar":
    #     return self.logical_not()

    # def setUb(self, ub):
    #     self.set_ub(ub)

    # def setLb(self, lb):
    #     self.set_lb(lb)

    # def setBounds(self, lb, ub):
    #     self.setLb(lb)
    #     self.setUb(ub)

    # def setValue(self, value):
    #     self.setBounds(value, value)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def X(self):
        return self.solution_value

    # @property
    # def VarIndex(self):
    #     return self.index

    @property
    def VarName(self) -> str:
        return self.get_name()  # type: ignore

    @property
    def LB(self):  # type: ignore
        return self.lb

    @property
    def UB(self):  # type: ignore
        return self.ub


DVar.X = CPlexVar.X  # type: ignore
DVar.VarName = CPlexVar.VarName  # type: ignore
# DVar.VarIndex = CPlexVar.VarIndex  # type: ignore
DVar.LB = CPlexVar.LB  # type: ignore
DVar.UB = CPlexVar.UB  # type: ignore
# DVar.setValue = CPlexVar.setValue  # type: ignore
# DVar.setLb = CPlexVar.setLb  # type: ignore
# DVar.setUb = CPlexVar.setUb  # type: ignore
# DVar.setBounds = CPlexVar.setBounds  # type: ignore
# DVar.Not = CPlexVar.Not  # type: ignore

DecisionKPI.X = CPlexVar.X  # type: ignore


class CPlexCpoVar(CpoIntVar, IVar):
    def __init__(
        self,
        is_integer_var: bool,
        solver,
        min=None,
        max=None,
        name=None,
        domain=None,
    ):
        if is_integer_var:
            super().__init__(_build_int_var_domain(min, max, domain), name)
        else:
            super().__init__((0, 1), name)
        self._solver = solver

    # def Not(self) -> "IVar":
    #     raise RuntimeError("NotImplemented")

    # def setUb(self, ub):
    #     raise RuntimeError("NotImplemented")

    # def setLb(self, lb):
    #     raise RuntimeError("NotImplemented")

    # def setBounds(self, lb, ub):
    #     raise RuntimeError("NotImplemented")

    # def setValue(self, value):
    #     raise RuntimeError("NotImplemented")

    @property
    def X(self):
        return self._solver.valueExpression(self)

    # @property
    # def VarIndex(self):
    #     raise RuntimeError("NotImplemented")

    @property
    def VarName(self) -> str:
        return self.get_name()

    @property
    def LB(self):
        return self.lb

    @property
    def UB(self):
        return self.ub



# -*- coding: utf-8 -*-
# @Time    : 2023/4/12 9:12
# @Author  : luyi
"""
约束规划
"""

from collections import defaultdict
from typing import Any, Optional, Union, List, Dict, Tuple
from typing_extensions import override
from .constants import Vtype, ObjType, OptimizationStatus, CmpType
from .interface_ import IVar, IModel
from .tupledict import tupledict
from .utilx import (
    INFINITY,
    is_list_or_tuple,
    get_combinations,
    is_bool_var,
)

from .variable_cplex import CPlexVar
from .func import name_str
from docplex.mp.model import Model
from docplex.mp.sdetails import SolveDetails
from docplex.mp.solution import SolveSolution
from docplex.mp.model_reader import ModelReader

Real = Union[float, int]


class CplexMpSolver(IModel):

    def __init__(self, name=""):
        super().__init__("CPLEX", name)
        self.__model = Model(self.name)
        self.__line_expr_object = defaultdict(list)
        self.__expr_simple_object = None

    @override
    def Sum(self, expr_array) -> "IVar":  # type: ignore
        return self.__model.sum(expr_array)  # type: ignore

    @override
    def setHint(self, start: Dict[IVar, Real]):
        warmstart = self.__model.new_solution()
        for var, value in start.items():
            warmstart.add_var_value(var, value)
        self.__model.add_mip_start(warmstart)

    @override
    def setTimeLimit(self, time_limit_seconds: int):
        self.__model.set_time_limit(time_limit_seconds)

    @override
    def wall_time(self) -> int:
        solver_details: SolveDetails = self.__model.solve_details  # type: ignore
        return solver_details.time * 1000

    @override
    def iterations(self) -> Real:
        solver_details: SolveDetails = self.__model.solve_details  # type: ignore
        return solver_details.nb_iterations

    @override
    def nodes(self) -> Real:
        solver_details: SolveDetails = self.__model.solve_details  # type: ignore
        return solver_details.nb_nodes_processed

    @override
    def addVar(
        self,
        lb: Real = 0.0,
        ub: Real = INFINITY,
        vtype: Vtype = Vtype.CONTINUOUS,
        name: str = "",
    ) -> IVar:
        """
            Add a decision variable to a model.
        :param lb: Lower bound for new variable.
        :param ub: Upper bound for new variable.
        :param vtype: variable type for new variable(Vtype.CONTINUOUS, Vtype.BINARY, Vtype.INTEGER).
        :param name: Name for new variable.
        :return: variable.
        """

        if vtype == Vtype.CONTINUOUS:
            var1: CPlexVar = self.__model.continuous_var(
                lb=lb, ub=ub, name=name
            )  # type: ignore
        elif vtype == Vtype.INTEGER:
            var1 = self.__model.integer_var(lb=lb, ub=ub, name=name)  # type: ignore
        else:
            var1 = self.__model.binary_var(name=name)  # type: ignore
        var1._solver = self  # type: ignore
        return var1

    @override
    def addVars(
        self,
        *indices,
        lb: Real = 0.0,
        ub: Real = INFINITY,
        vtype: Vtype = Vtype.CONTINUOUS,
        name: str = "",
    ) -> tupledict:
        li = []
        for ind in indices:
            if isinstance(ind, int):
                ind = [i for i in range(ind)]
            elif is_list_or_tuple(ind):
                pass
            elif isinstance(ind, str):
                ind = [ind]
            else:
                raise ValueError("error input")
            li.append(ind)
        all_keys_tuple = get_combinations(li)
        tu_dict = tupledict(
            [
                [key, self.addVar(lb, ub, vtype, name_str(name, key))]
                for key in all_keys_tuple
            ]
        )
        return tu_dict

    @override
    def getVars(self) -> List[IVar]:
        raise RuntimeError("not implemented")

    @override
    def getVar(self, i) -> IVar:
        return self.__model.get_var_by_index(i)  # type: ignore

    @override
    def addConstr(self, lin_expr):
        self.__model.add_constraint_(lin_expr)

    @override
    def addConstrs(self, lin_exprs):
        # 检查下是否可以迭代。
        self.__model.add_constraints_(lin_exprs)

    @override
    def setObjective(self, expr):
        self.__expr_simple_object = expr

    @override
    def setObjectiveN(
        self, expr, index: int, priority=None, weight: float = 1, name: str = ""
    ):
        """

        :param expr:
        :param index:
        :param priority:
        :param weight:
        :param name:
        :return:
        """
        self.__line_expr_object["exprs"].append(expr)
        self.__line_expr_object["priority"].append(priority)
        self.__line_expr_object["weight"].append(weight)

    @override
    def addGenConstrAnd(self, resvar, varList: List[IVar]):
        """
        和 addGenConstrAnd(y, [x1,x2])
        :param resvar:
        :param varList:
        :return:
        """

        self.addConstr(resvar == self.__model.logical_and(*varList))

    @override
    def addGenConstrOr(self, resvar: IVar, varList: List[IVar]):
        self.addConstr(resvar == self.__model.logical_or(*varList))

    @override
    def addGenConstrXOr(self, resvar: IVar, varList: List[IVar]):
        super().addGenConstrXOr(resvar, varList)

    @override
    def addGenConstrPWL(
        self, var_x: IVar, var_y: IVar, x_range: List, y_range: List, M
    ):
        super().addGenConstrPWL(var_x, var_y, x_range, y_range, M)

    @override
    def addGenConstrIndicator(
        self,
        binvar: IVar,
        binval: bool,
        lhs: IVar,
        sense: CmpType,
        rhs: float,
        M: float,
    ):
        """
        若 binvar 为binval ,则 lhs 与 rhs 之间有sense 的关系

        :param binvar: 0-1变量
        :param binval: bool 常量
        :param lhs:  左侧变量
        :param sense: 等号，大于等于，小于等于
        :param rhs: 右侧常量
        :param M: 大M
        :return:
        """
        super().addGenConstrIndicator(binvar, binval, lhs, sense, rhs, M)

    @override
    def addIndicator(self, binvar: IVar, binval: bool, constr):
        self.__model.add_indicator(binvar, constr, binval)

    @override
    def addGenConstrAbs(self, resvar, var_abs: IVar, M):
        self.addConstr(self.__model.abs(var_abs) == resvar)

    @override
    def addGenConstrMultiply(self, z: IVar, l: Tuple[IVar, IVar]):
        """x * y = z"""
        super().addGenConstrMultiply(z, l)
        x = l[0]
        y = l[1]
        if not is_bool_var(x) and not is_bool_var(y):
            raise RuntimeError("At least one binary variable is required.")
        if is_bool_var(y):
            x, y = y, x
        M = y.UB
        self.addConstr(z <= y)
        self.addConstr(z <= x * M)
        self.addConstr(z >= y + (x - 1) * M)

    @override
    def addRange(
        self, expr, min_value: Union[float, int], max_value: Union[float, int]
    ):
        self.__model.add_range(min_value, expr, max_value)

    @override
    def calc_lbub_and_expr(self, constr) -> Tuple[Real, Real, Any]:
        expr = constr.left_expr - constr.right_expr
        sign = constr.cplex_code  # 'L' (≤), 'G' (≥), 'E' (=)
        lb = -self.INFINITY
        ub = self.INFINITY
        if sign == "L":
            ub = 0
        elif sign == "G":
            lb = 0
        elif sign == "E":
            lb = 0
            ub = 0
        return lb, ub, expr

    @override
    def addConstrOr(
        self,
        constrs: List,
        ok_num: int = 1,
        cmp_type: CmpType = CmpType.EQUAL,
        M=None,
    ):
        """
        约束的满足情况

        :param constr: 所有的约束
        :param ok_num: 需要满足的个数，具体则根据cmpType
        :param cmpType: CmpType.LESS_EQUAL CmpType.EQUAL,CmpType.GREATER_EQUAL
        :return:
        """
        constr_num = len(constrs)
        x = []
        for i in range(constr_num):
            x.append(self.addVar(vtype=Vtype.BINARY))
            constr = constrs[i]
            lb, ub, expr = self.calc_lbub_and_expr(constr)
            tempM = self.INFINITY
            if M is not None:
                tempM = M
            if lb > -self.INFINITY:  # 若大于
                self.addConstr(expr + tempM * (1 - x[i]) >= lb)
            if ub < self.INFINITY:
                self.addConstr(expr - tempM * (1 - x[i]) <= ub)

        if cmp_type == CmpType.EQUAL:
            self.addConstr(self.Sum(x) == ok_num)
        elif cmp_type == CmpType.GREATER_EQUAL:
            self.addConstr(self.Sum(x) >= ok_num)
        elif cmp_type == CmpType.LESS_EQUAL:
            self.addConstr(self.Sum(x) <= ok_num)  # type: ignore
        else:
            raise Exception("error value of cmpType")

    @override
    def numVars(self, vtype: Optional[Vtype] = None) -> int:
        if vtype is None:
            return self.__model.number_of_variables
        elif vtype == Vtype.CONTINUOUS:
            return self.__model.number_of_continuous_variables
        elif vtype == Vtype.INTEGER:
            return self.__model.number_of_integer_variables
        else:
            return self.__model.number_of_binary_variables

    @override
    def numConstraints(self) -> int:
        return self.__model.number_of_constraints

    @override
    def write(self, filename: str, obfuscated=False):
        filename = filename.lower()
        if filename.endswith(".lp"):
            self.__model.export_as_lp(filename)
        elif filename.endswith(".mps"):
            self.__model.export_as_mps(filename)
        elif filename.endswith(".proto"):
            raise TypeError(".proto 导出异常，待修复")

    @override
    def read(self, path: str):
        self.clear()
        m = ModelReader.read_model(path)
        self.__model = m

    @property
    @override
    def ObjVal(self) -> Real:
        return sum(self.__model.multi_objective_values)

    @property
    def has_cplex_env(self):
        return self.__model.environment.has_cplex

    @override
    def optimize(self, obj_type: ObjType = ObjType.MINIMIZE) -> OptimizationStatus:
        if self._params.TimeLimit:
            self.setTimeLimit(self._params.TimeLimit)
        if self._params.MIPGap:
            self.__model.parameters.mip.tolerances.mipgap = self._params.MIPGap  # type: ignore
        if self._params.Precision:
            self.__model.float_precision = self._params.Precision
        if self._flag_objective_n:
            self.__model.set_multi_objective(
                "min" if obj_type is ObjType.MINIMIZE else "max",
                self.__line_expr_object["exprs"],
                self.__line_expr_object["priority"],
                self.__line_expr_object["weight"],
            )
        else:
            self.__model.set_objective(
                "min" if obj_type is ObjType.MINIMIZE else "max",
                self.__expr_simple_object,
            )
        status: SolveSolution = self.__model.solve(
            log_output=self._params.EnableOutput
        )  # type: ignore

        if status:
            if status.is_feasible_solution or status.is_valid_solution:
                result = OptimizationStatus.OPTIMAL
            else:
                result = OptimizationStatus.FEASIBLE
        else:
            result = OptimizationStatus.ERROR
        return result

    @override
    def clear(self):
        self.__model.clear()

    @override
    def close(self):
        self.__model.end()

    # 独有。

    @override
    def addNoOverlap(self, interval_vars, M):
        """
        相邻不重复。

        :param interval_vars:
        :return:
        """
        raise RuntimeError("not implemented")

    @override
    def newIntervalVar(self, start, size, end):
        raise RuntimeError("not implemented")

    @override
    def valueExpression(self, expression):
        return expression.solution_value

    @override
    def setNumThreads(self, num_theads: int):
        self.__model.context.cplex_parameters.threads = num_theads
